/*
 * Copyright (C) 2014 digitalSTROM.org, Schlieren, Switzerland
 *
 * digitalSTROM ds485-client library is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * digitalSTROM ds485-client library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with digitalSTROM ds485-client library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "socket_client.h"

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include "log.h"
#include "ds485-client-interface.h" /* TODO error codes */

#ifndef SOCK_CLOEXEC
#define SOCK_CLOEXEC 0
#endif

int get_local_port(int socket, int *port)
{
    struct sockaddr_in localAddress;
    socklen_t addressLength = sizeof(localAddress);
    int ret;

    if (port == NULL) {
        return -EINVAL;
    }
    if (socket < 0) {
        return -EBADF;
    }
    ret = getsockname(socket, (struct sockaddr *)&localAddress, &addressLength);
    if (ret < 0) {
        return -errno;
    }
    *port = ntohs(localAddress.sin_port);
    return 0;
}

/**
 * socket_connect -- connect to service on host, until success or abort is
 * false
 * @host - hostname or "127.0.0.1" notation
 * @service - 'telnet' or port number
 * @protocol - "tcp" / "udp"
 * @cancel - will abort retry if cancel is true
 */
int socket_connect(char *host, char *service, char *protocol, bool *cancel)
{
    struct hostent *phe;
    struct servent *pse;
    struct protoent *ppe;
    struct sockaddr_in sin;
    int s, type, proto, opts, one = 1;
    struct pollfd fds[1];

    memset((char *) &sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;

    if ((pse = getservbyname(service, protocol)))
        sin.sin_port = htons(ntohs((unsigned short) pse->s_port));
    else if ((sin.sin_port = htons((unsigned short) atoi(service))) == 0)
        return ERROR_NO_PORT_FOUND;

    if ((phe = gethostbyname(host)))
        memcpy((char *) &sin.sin_addr, phe->h_addr, phe->h_length);
    else if ((sin.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)
        return ERROR_INVALID_ADDRESS;

    ppe = getprotobyname(protocol);
    if (strcmp(protocol, "udp") == 0) {
        type = SOCK_DGRAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_UDP;
    }
    else {
        type = SOCK_STREAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_TCP;
    }

    if ((s = socket(PF_INET, SOCK_CLOEXEC | type, proto)) < 0)
        return ERROR_CREATE_SOCKET;

#ifdef __APPLE__
    fcntl(s, F_SETFD, FD_CLOEXEC);
#endif

    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof(one)) == -1) {
        close(s);
        return ERROR_SOCKET_OPTIONS;
    }

    opts = fcntl(s, F_GETFL);
    if (opts > 0) {
        fcntl(s, F_SETFL, opts | O_NONBLOCK);
    }

    int ret = connect(s, (struct sockaddr *) &sin, sizeof(sin));

    if (ret != 0) {
        if (errno != EINPROGRESS && errno != EALREADY) {
            perror("connectsock connect:");
            close(s);
            return ERROR_SOCKET_CONNECT;
        }

        memset(&fds, 0, sizeof(fds));
        fds[0].fd = s;
        fds[0].events = POLLOUT;

        /* socket is non-blocking and we have to poll for connect completion.
         *
         * Non-blocking mode is used to work around very long (>180s) timeouts
         * that result in the process not being interruptible if there is no
         * server available. */
        int retries = 60; /* 1 min */
        while (!(*cancel) && (retries-- > 0)) {
            ret = poll(fds, 1, 1000);
            if (ret == 0) {
                /* timeout */
                continue;
            } else if (ret < 0) {
                perror("connectsock select:");
                close(s);
                return ERROR_SOCKET_CONNECT;
            }

            /* see "man connect" EINPROGRESS */
            int err = -1;
            int l_sock_optval_len = sizeof(err);
            if (getsockopt(s, SOL_SOCKET, SO_ERROR, (int*)&err,
                           (socklen_t*)&l_sock_optval_len) != 0) {
                perror("connectsock getsockopt:");
                close(s);
                return ERROR_SOCKET_CONNECT;
            }

            if (err != 0) {
                fprintf(stderr, "connectsock einprogress: %s\n", strerror(err));
                close(s);
                return ERROR_SOCKET_CONNECT;
            }

            break;
        }
    }

    /* return to blocking mode */
    opts = fcntl(s, F_GETFL);
    if (opts > 0) {
        fcntl(s, F_SETFL, opts & ~O_NONBLOCK);
    }

#ifdef IPTOS_LOWDELAY
    opts = IPTOS_LOWDELAY;
    setsockopt(s, IPPROTO_IP, IP_TOS, &opts, sizeof(opts));
#endif

#ifdef TCP_NODELAY
    if (type == SOCK_STREAM)
        setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));
#endif
    return s;
}
